/* ****************************************** */
/* Emu75: HP-75  software emulator for PC     */
/*                                            */
/* emu75.c   main module - PC version         */
/*                                            */
/* Version 1.3  J-F Garnier 2010-2018         */
/* ****************************************** */


/* Copyright (c) 2010-2018  J-F Garnier
'
' This program is free software; you can redistribute it and/or
' modify it under the terms of the GNU General Public License
' as published by the Free Software Foundation; either version 2
' of the License, or (at your option) any later version.
'
' This program is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with this program; if not, write to the Free Software
' Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
'
' ------------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <conio.h>
#include <alloc.h>
#include <time.h>



/* definition of UNIT_TICKS for clock() units */
#if __TURBOC__  >= 0x400
#define UNIT_TICKS 1
#else
#define UNIT_TICKS 55
#endif

extern unsigned _stklen = 8000;


#define GLOBAL
#include "capricor.h"

#include "version.h"

extern int commande(void);  /* monit.c */

extern void update_display(char *buf1, char *buf2, int cpos, int ctype); // userpc.c


char SYSROM75[40]= "rom75.bin";
char BASROM75[40]= "";

#define RAMSIZE 24

//unsigned char buffer[2048];    /* temporary buffer - moved to capricorn.h */

char f_trace;   /* trace */

int fmonit;

/* variables module description  */
struct portdevice {
  int nuport;      /* port */
  int type;        /* module type : A=RAM, O=ROM, R=HRD */
  int taille;      /* module size (kB)*/
  int sltaddr;     /* slot address */
  char nomfic[40]; /* file associated with the module */
  int seg;	   /* segment holding the module */
} ;

struct romslot {
  int sltaddr;     /* slot address */
  int seg;	   /* segment holding the ROM */
  int fwrit;       /* writeable flag */
} ;

struct portdevice device75[20]; /* 19 max worst case */
struct romslot rombank[20];     /* 19 banked ROM 75 max */

// slot address - not used yet
int slot75[24] = { 0xFF46, 0xFF4e, 0xFF36, 0xFF3E,
                   0xFF40, 0xFF48, 0xFF30, 0xFF38,
                   0xFF42, 0xFF4A, 0xFF32, 0xFF3A,
                   0xFF44, 0xFF4C, 0xFF34, 0xFF3C,
                   0xFF50, 0xFF52, 0xFF54, 0xFF56,
                   0xFF58, 0xFF5A, 0xFF5C, 0xFF5E };

int speed_emu;  /* number of instructions emulated per tick */


/* for module desas.c : */
char from=1; /* changed JFG 17/05/07 */
char f_xil=0;

int nbdev;  /* number of devices */
int tabdev[32];

int flowpower;
int fdiag;  // diagnostic ROM present   v1.1

#if __TURBOC__ < 0x200
void _fmemcpy(int far *dest, int far *src, int n)
{
  n /= 2;
  while (n--) *dest++ = *src++;
}
void _fmemset(int far *dest, int val, int n)
{
  n /= 2;
  while (n--) *dest++ = val;
}
#endif



/* macro to build a pointer into Emu75 memory space. */
#define MK_FP75(a) MK_FP(tab_seg[a>>13],(int)(a&0x1fff))




/* ************************************************************* */
/* keyboard management  					 */
/* Emu75 emulates the keyboard at key buffer level,              */
/* not at register level.                                        */
/* ************************************************************* */

/*
=SVCWRD	DAD	8284
=KEYHIT	DAD	835F
*/


/* ****************************************** */
/* keybufst()                                 */
/*                                            */
/* read the key buffer status                 */
/* 0: KEYHIT free, 1: KEYHIT not free         */
/* ****************************************** */
int keybufst(void)
{
  unsigned char far *ptr;
  int n;

  ptr =MK_FP75(0x8284);          /* svcwrd */

  n = (*ptr)&1;
  return(n);
}



/* ****************************************** */
/* push_key()                                 */
/*                                            */
/* push a key int the key buffer              */
/* ****************************************** */
int push_key(int key)
{
  unsigned char far *ptr;

  ptr =MK_FP75(0x8284);          /* svcwrd */
  if ( (key==0x80) || !(*ptr&1) ) {
    /* ATTN or KEYHIT free: put key */
    *ptr |= 1;
    ptr =MK_FP75(0x835F);          /* keyhit */
    *ptr = key;
    return(0);
  }
  else
    return(1);
}



/* ****************************************** */
/* toPhyKey()                                 */
/*                                            */
/* translates an ASCII keycode                */
/*    to a physical keycode (for diag rom)    */
/* doesn't manage shift/ctrl                  */
/* ****************************************** */
static int toPhyKey(int k)
{
  int c;
  static char tab_kb[]= {
    /* 42-47 * + , - . /  */
    0x23, 0x2F, 0x3D, 0x21, 0x08, 0x03,
    /* 48-63 0-9, ; =  */
    0x29, 0x68, 0x63, 0x6B, 0x67, 0x6F, 0x61, 0x69,
    0x65, 0x6D,    0, 0x13,    0, 0x1B,    0,    0,
    /* 64-95 A-Z */
       0, 0x43, 0x31, 0x37, 0x47, 0x5B, 0x4F, 0x41,
    0x49, 0x55, 0x45, 0x4D, 0x18, 0x35, 0x39, 0x5D,
    0x28, 0x58, 0x57, 0x4B, 0x5F, 0x59, 0x3F, 0x53,
    0x3B, 0x51, 0x33,    0,    0,    0,    0,    0,
    /* 96-127 a-z */
       0, 0x43, 0x31, 0x37, 0x47, 0x5B, 0x4F, 0x41,
    0x49, 0x55, 0x45, 0x4D, 0x18, 0x35, 0x39, 0x5D,
    0x28, 0x58, 0x57, 0x4B, 0x5F, 0x59, 0x3F, 0x53,
    0x3B, 0x51, 0x33,    0,    0,    0,    0,    0,
    /* 128-143 special keys */
    0x70, 0x78, 0x73, 0x7B, 0x77, 0x7F, 0x71, 0x79,
    0x75, 0x7D, 0x2D, 0x25, 0x48, 0x0B, 0x2B, 0x27 } ;

  if (k==8)
    c=0x27;
  else if (k==13)
    c=0x17;
  else if (k==' ')
    c=0x07;
  else if ((k>=42)&&(k<=143))
    c=tab_kb[k-42];
  else
    c=0;

  if (c) {
    c= (c*2)+1;
    return(c);
  }
  else
    return(0);
}



/* ****************************************** */
/* traite_touche()                            */
/*                                            */
/* manage the keyboard                        */
/* flag fwait means: wait for a key           */
/* 0.96 use pending key                       */
/* 1.1  manage physical key for diag75        */
/* ****************************************** */
int static traite_touche(int fwait)
{
  int k;
  static int pk=0; // pending key
  unsigned char far *ptr;


  k=readkey(fwait);

  // v1.1
  if (fdiag) {
    if (k>0) {
      // put physical key into regKey
      k=toPhyKey(k);
      if (k) {
        regKey=k;
        ptr =MK_FP75(0xFF02);  /* key status reg */
        *ptr |= 0x2c;  // key available, shift&ctrl inactive
      }
    }
    return(k);
  }

  if (k>0) pk=k;

  if (pk>0) {
    // push key if intr enabled
    if (fintr) {
      push_key(pk);
      pk=0;
    }
  }

  return(k);
}



/* ************************************************************* */
/* display management                                   	 */
/* Emu75 emulates the LCD at display buffer level,               */
/* not at LCD controller level.                                  */
/* ************************************************************* */

/*
=CREST	DAD	82DA
=I/RFLG	DAD	82DB
=CURSOR	DAD	82DC
=LETSEE	DAD	82DD
=LCDESC	DAD	82DE
=JUSTSD	DAD	82DF
=LCDOFF	DAD	82E0
=DEAD	DAD	82E1
=LCDPTR	DAD	82E2    buffer position of cur.
=LCDWIN	DAD	82E4
=LCDBUF	DAD	82E6    96 char.
=GL.LIN	DAD	8346
=GL.CUR	DAD	8348
=GL.LEN	DAD	834A
=ANNFLG	DAD	834B
=SHIELD	DAD	834C
*/




/* ****************************************** */
/* traite_display()                           */
/*                                            */
/* manage the LCD display                     */
/* flag frun means: CPU running ('*' announc) */
/* ****************************************** */
void static traite_display(int frun)
{
  static char derligne[98] = "";
  static char derflags[8] = "";
  static char derligne2[34]="";
  static int derpos=-1;
  static int deretat=-1;
  static int derdeb=0;
  char ligne[98];
  char ligne2[130];
  char flags[8];
  unsigned char far *ptr;
  int xp, xs, deb, ofs;
  int faff, fcsr, flg;

  faff=0; fcsr=0; flg=0;

  if (fdiag) {
    // v1.1 for the diag ROM
    strncpy(ligne2,regLCD,32);
    ligne2[32]=0;
    if (strcmp(ligne2,derligne2)!=0) {
      strcpy(derligne2,ligne2);
      faff=1;
    }
    flags[0]= 0;
    flags[4]=frun;
    if (memcmp(flags,derflags,5)!=0) {
      memcpy(derflags,flags,5);
      flg=1;
    }
    if (faff||flg)
      update_display(ligne2, flags, 1, 0x40);
    return;
  }

  ptr =MK_FP75(0x82e6);  /* LCDBUF display buffer */

  _fmemcpy(ligne,ptr,96);
  ligne[96]=0;
  if (strcmp(ligne,derligne)!=0) {
    strcpy(derligne,ligne);
    faff=1;
    derdeb=0;
  }

  // 0.95: WINDOW management (for Visicalc mainly, but also I/O ROM)
  ptr=MK_FP75(0x8450);   /* MINMIN */
  ofs = 80 - (*ptr);   // offset
  if (ofs<0) ofs=0;
  if (ofs>32) ofs=32;
  ligne2[0]=0;
  if (ofs>0) {
    strncpy(ligne2,regLCD,32);
    ligne2[ofs]=0;
    if (strcmp(ligne2,derligne2)!=0) {
      strcpy(derligne2,ligne2);
      faff=1;
    }
  }

  /* cursor position */
  ptr = MK_FP75(0x82e2);  /* LCDPTR */
  xp= (*ptr) + ((*(ptr+1))<<8) - 0x82e6;
  if (xp<0) xp=0;
  if (xp>95) xp=95;
  if ((xp!=derpos)||faff) {
    deb=derdeb;
    if ((xp<(78-ofs)/2))
      deb=0;
    else if (xp<deb)
      deb=xp;
    else if (xp>(deb+78-ofs))
      deb=xp-78+ofs;
    derpos=xp;
    if (deb!=derdeb) {
      derdeb=deb;
      faff=1;
    }
    fcsr=1;
  }
  else
    deb=derdeb;

  if (faff) {
    if (derdeb<(96-79))
      ligne[deb+79]=0;
    fcsr=1;
  }
  /* cursor status */
  ptr = MK_FP75(0x82dc);  /* CURSOR */
  if (*ptr) {
    // cursor on
    ptr = MK_FP75(0x82db);  /* I/RFLG	*/
    if (*ptr)
      xs=0x80; // insert
    else
      xs=0;    // replace
  }
  else
    xs=0x40;  // no cursor
  // xs codes have been chosen in line with emu71 code
  if (xs!=deretat) {
    fcsr=1;
    deretat=xs;
  }

  /* flags */
  ptr=MK_FP75(0x834B);   /* ANNFLG */
  flags[0]= *ptr;
  flags[4]=frun;
  if (memcmp(flags,derflags,5)!=0) {
    memcpy(derflags,flags,5);
    flg=1;
  }

  if (faff) {
    if (ofs>0) {
      strcat(ligne2,&ligne[deb]);
      update_display(ligne2, NULL, xp-deb+1+ofs, xs);
    }
    else
      update_display(&ligne[deb], NULL, xp-deb+1, xs);
    fcsr=0;
  }
  if (fcsr) {
    update_display(NULL, NULL, xp-deb+1+ofs, xs);
  }
  if (flg) {
    update_display(NULL, flags, xp-deb+1+ofs, xs);
  }

}



/* ************************************************************* */
/* timer management             				 */
/* ************************************************************* */






/* ****************************************** */
/* incr_timer()                               */
/*                                            */
/* increment the HP75 timer                   */
/* ****************************************** */
void static incr_timer(int n)
{
  timer_l += n & 255;
  if (timer_l>255) {
    timer_l -=256;
    timer++;
  }
  timer += n >> 8;
}




/* ****************************************** */
/* check_timer()                              */
/*                                            */
/*  check timer for interrupt                 */
/* ****************************************** */
int static check_timer(void)
{
  int res;

  res=0;
  if (timer>tcomp) {
    if ( fintr && (compstat&2) ) {
      do_intr(8);  // cause intr with vector at addr=8
      res=1;
    }
  }

  return(res);
}





/* ************************************************************* */
/* file loading 						 */
/* ************************************************************* */


/* ****************************************** */
/* loadrom()                                  */
/*                                            */
/*  load the mainframe ROM                    */
/* ****************************************** */
void loadrom(char *sysrom, char *basrom)
{
  FILE *fic;
  unsigned char far *ptr;
  int i, n;

  /* load memory */
  fic=fopen(sysrom,"rb");
  if (fic==NULL) {
    fin_ecran();
    fprintf(stderr,"Emu75: file %s not found\n",sysrom);
    exit(0);
  }
  ptr=MK_FP(tab_seg[0],0);
  for (i=0;i<32;i++) {
    n=fread(buffer,1,1024,fic);
    if (n!=1024) break;
    _fmemcpy(ptr,buffer,1024);
    ptr+=1024;
  }
  fclose(fic);
  if ((i!=24)&&(i!=32)) {
    fin_ecran();
    fprintf(stderr,"Emu75: error in file %s\n",sysrom);
    exit(0);
  }

  if (i==24) {
    fic=fopen(basrom,"rb");
    if (fic==NULL) {
      fin_ecran();
      fprintf(stderr,"Emu75: file %s not found\n",basrom);
      exit(0);
    }
    ptr=MK_FP(tab_seg[7],0);
    for (i=0;i<8;i++) {
      n=fread(buffer,1,1024,fic);
      if (n!=1024) break;
      _fmemcpy(ptr,buffer,1024);
      ptr+=1024;
    }
    fclose(fic);
    if (i!=8) {
      fin_ecran();
      fprintf(stderr,"Emu75: error in file %s\n",basrom);
      exit(0);
    }
  }

}


/* ****************************************** */
/* loadmodule()                               */
/*                                            */
/*  load a module ROM                         */
/* ****************************************** */
void static loadmodule(int seg, int taille, char *nomfic)

{
  FILE *fic;
  unsigned char far *ptr;
  int i, n;
  char buf[80];

  /* init memory */
  ptr=MK_FP(seg,0);
  for (i=0;i<taille;i++) {
    _fmemset(ptr,0,1024);
    ptr += 1024;
  }

  fic=fopen(nomfic,"rb");
  if (fic==NULL) {
    sprintf(buf,"Emu75: file %s not found\r\n",nomfic);
    disp_string(buf);
  }
  else {
    ptr=MK_FP(seg,0);
    for (i=0;i<taille;i++) {
      n=fread(buffer,1,1024,fic);
      if (n!=1024) break;
      _fmemcpy(ptr,buffer,1024);
      ptr+=1024;
    }
    fclose(fic);
    if (i!=taille) {
      fin_ecran();
      fprintf(stderr,"Emu75: error in file %s\n",nomfic);
      exit(0);
    }
  }
}






/* ************************************************************* */
/* initialization						 */
/* ************************************************************* */


/* ****************************************** */
/* init_register()                            */
/*                                            */
/*   init the CPU registers                   */
/* ****************************************** */
void init_register(void)
{
  unsigned char far *ptr;
  int i;

  for (i=0;i<64;i++) regCPU.byt[i]=0;
  regARP=0; regDRP=0;
  regE=0; regFlags=0;
  regROMslt=0;

  // set PC to reset vector
  regCPU.reg.pc = lit2(0);

  // init I/O
  ptr=MK_FP75(0xff82);
  *ptr = 0;  // power on reset

  fintr=0;  // GIE=0 global interrupt disabled


}


/* ****************************************** */
/* init_ram()                                 */
/*                                            */
/*   init the RAM                             */
/* ****************************************** */
void init_ram(void)
{
  unsigned char far *ptr;
  int i;

  /* initialize RAM to blank */
  ptr=MK_FP(tab_seg[4],0);
  for (i=0;i<24;i++) {
    _fmemset(ptr,32,1024);
    ptr+=1024;
  }


}



/* ****************************************** */
/* maj_date()                                 */
/*                                            */
/*  set the date & time in Emu75              */
/* ****************************************** */
void static maj_date(void)
{
  int i, n;
  unsigned char far *ptr;
  struct date d;
  struct time t;
  long ld, lt;

  getdate(&d);

  // number of days since 1/1/2000 and 1/1/current year:
  n = (d.da_year)%100;
  n = n*365 + (n-1)/4;  // one extra day every 4 years

  for (i=1;i<d.da_mon;i++) {
    switch (i) {
      case 2:
	if (d.da_year%4) n += 28; else n += 29;  /* warning: will fail in 2100 ! */
	break;
      case 4:
      case 6:
      case 9:
      case 11:
	n += 30;
	break;
      default:
	n += 31;
    }
  }
  n += d.da_day;


  gettime(&t);
  lt=t.ti_hour;
  lt *=60;
  lt += t.ti_min;
  lt *= 60;
  lt += t.ti_sec;

  lt +=n * 86400L;  // in seconds


  // calc t1=TMBASE + timer in 2^4s
  //      t2=get_dt()+magic; in 2^4s
  // calc  t2-t1, add to timer
  // timer = timer + get_dt()+magic-(TMBASE+timer)
  // timer = get_dt()+magic-TMBASE
  // 01/01/2000 00h:
  // TMBASE (8394) = xx xx DF 6F 78 AC 03
  // 03AC786FDF in 4s units

  ptr=MK_FP75(0x8394);  // TMBASE
  ptr +=2; // point on part in 2^4s unit
  ld = * ( (long *) ptr);
  ld = 0xAC786FE0L -ld;
  ld += lt>>2;

  timer = ld << 8;
  timer_l=0;

}






/* ****************************************** */
/* loaddevice()                               */
/*                                            */
/*  load and init a HPIL device               */
/* ****************************************** */
void static loaddevice(char *nomdev, char *option)
{
  int tdev;
  int d, n;

  if      (stricmp(nomdev,"DISPLAY")==0) tdev=1;
  else if (stricmp(nomdev,"HDRIVE1")==0) tdev=2;
  else if (stricmp(nomdev,"FDRIVE1")==0) tdev=3;
  else if (stricmp(nomdev,"DOSLINK")==0) tdev=4;
  else if (stricmp(nomdev,"XIL")==0) {
    if ((option[0]=='c')||(option[0]=='C'))
     tdev=10;
    else
     tdev=5;
  }
  else if (stricmp(nomdev,"LPRTER1")==0) tdev=6;
  else if (stricmp(nomdev,"HDRIVE2")==0) tdev=7;
  else if (stricmp(nomdev,"SERIAL1")==0) tdev=8;
  else if (stricmp(nomdev,"SERIAL2")==0) tdev=9;
  /* XIL COMx is tdev=10 */
  else tdev=0;

  if (tdev && (nbdev<32)) {
    tabdev[nbdev]=tdev;
    nbdev++;
  }
  switch (tdev) {
    case 1:
      init_ildisplay();
      break;
    case 2:
      init_ilhdisc(option);
      break;
    case 3:
      if ((option[0]=='b')||(option[0]=='B')) d=1; else d=0;
      init_ilfdisc(d);
      break;
    case 5:
#ifdef XIL
      d=0;
      if (option[0]!=0) sscanf(option,"%x",&d);
      if (d==0) d=0x1700;
      init_ilext(d,1);
      f_xil=1;
#endif
      break;
    case 6:
      init_ilcentro();
      break;
    case 7:
      init_ilhdisc2(option);
      break;
    case 8:
      d=0;
      n=0xe3;
      if (option[0]!=0) {
        if (option[1]==0) {
          sscanf(option,"%d",&d);
          d--;
        }
        else if (option[1]==',') {
          sscanf(option,"%d,%x",&d,&n);
          d--;
        }
        else {
          sscanf(option,"%x",&n);
        }
      }
      init_ilserial1(d,n);
      break;
    case 9:
      d=1;
      n=0xe3;
      if (option[0]!=0) {
        if (option[1]==0) {
          sscanf(option,"%d",&d);
          d--;
        }
        else if (option[1]==',') {
          sscanf(option,"%d,%x",&d,&n);
          d--;
        }
        else {
          sscanf(option,"%x",&n);
        }
      }
      init_ilserial2(d,n);
      break;
    case 10:  /* XIL2 */
      if ((option[1]!=0)&&(option[2]!=0)&&(option[3]!=0)) {
        sscanf(&option[3],"%d",&d);
        init_ilext2(d,1);
      }
      break;

  }

}




void selectbank(int slot)
{
  int i;

  tab_seg[3]=0;
  fwrom=0;
  if (slot>0) {
    for (i=0;i<20;i++) {
      if (rombank[i].seg==0) break;
      if (rombank[i].sltaddr==slot) {
        tab_seg[3]=rombank[i].seg;
        fwrom=rombank[i].fwrit;
        break;
      }
    }
  }
}



/* ****************************************** */
/* init_emulate()                             */
/*                                            */
/* read the file EMU75.INI                    */
/* initialize structures device and tab_seg   */
/* ****************************************** */
void static init_emulate(int dbg)
{
  unsigned char far *ptrmem;
  int i, n, nbre, seg;
  FILE *fic;
  char ligne[82];
  int nuport, taille;
  char nomfic[40];
  char type[8];
  char nomdev[40];
  char option[40];
  char f;

  for (i=0;i<8;i++) {
    tab_seg[i]=0;
    xmem_seg[i]=0;
  }
  for (i=0;i<20;i++) {
    device75[i].type=0;
    device75[i].sltaddr=0;
    rombank[i].seg=0;
    rombank[i].sltaddr=0;
    rombank[i].fwrit=0;
  }
  instmemic=0;  // default no MEMIC
  xmemsize=32;
  strcpy(xmemfile,"podmem.dat");


  /* read file .ini, and fill struct device */
  f=1;
  fic=fopen("emu75.ini","rt");
  if (fic==NULL) {
    fin_ecran();
    printf("Emu75: file emu75.ini not found !\n");
    exit(0);
  }
  else {
    nbre=0;
    while (fgets(ligne,80,fic)) {
      if ( (ligne[0]!=';') && (ligne[0]!=' ') && (ligne[0]!='\n') ) {
        if (ligne[0]=='[') {
          if (strnicmp(&ligne[1],"MOD",3)==0)
            f=1;
          else if (strnicmp(&ligne[1],"DEV",3)==0)
            f=2;
          else if (strnicmp(&ligne[1],"OPT",3)==0)
            f=3;
          else if (strnicmp(&ligne[1],"END",3)==0)
            f=4;
        }
        else {
          switch (f) {
          case 1:  /* modules */
            n=sscanf(ligne,"%d %s %d %s",&nuport,type,&taille,nomfic);
            if (n==4) {
              if ((type[1]=='R')&&(nuport==0)) {
                /* internal ROM */
                if (taille>=24)
                  strcpy(SYSROM75,nomfic);  /* system ROM */
                else
                  strcpy(BASROM75,nomfic);  /* basic ROM */
              }
              else {
                if (nbre<18) {
                  // 19 max worst case
                  device75[nbre].nuport = nuport;
                  device75[nbre].type = type[1];
                  device75[nbre].taille=taille;
                  strcpy(device75[nbre].nomfic , nomfic);
                  nbre++;
                }
              }
            }
            break;
          case 2:  /* devices */
            option[0]=0;
            n=sscanf(ligne,"%s %s",nomdev,option);
            if (n>=1) {
              loaddevice(nomdev, option);
            }
            break;
          case 3: /* options */
            n=sscanf(ligne,"%s %d %d %s",option,&nuport,&taille,nomfic);
            if ((n>=2)&&(strnicmp(option,"MEMIC",5)==0)) {
              instmemic=nuport;
              if (n==4) {
                xmemsize=taille;
                strcpy(xmemfile,nomfic);
              }
            }
            break;
          }
        }
      }
    }

    fclose(fic);

  }


  /* memory allocation */

  /* --- internal ROM --- */

  /* try to allocate internal ROM in EMS to save memory */
  seg=GetEmsSegment();
  if (seg!=0) {
    tab_seg[0]=seg;
    if (dbg) fprintf(stderr,"Emu75: Using EMS\n");
  }
  else {
    if (dbg) fprintf(stderr,"Emu75: No EMS\n");
    ptrmem=farmalloc(0x08020L);
    if (ptrmem==NULL) {
      fin_ecran();
      fprintf(stderr,"Emu75: No enough memory\n");
      exit(0);
    }
    /* security */
    if (FP_OFF(ptrmem)>15) {
      fin_ecran();
      fprintf(stderr,"Emu75: invalid memory pointer\n");
      exit(0);
    }
    /* ajust for beginning of page */
    tab_seg[0]=FP_SEG(ptrmem)+1;
  }
  tab_seg[1]=tab_seg[0]+512;
  tab_seg[2]=tab_seg[1]+512;
  tab_seg[7]=tab_seg[2]+512;

  loadrom(SYSROM75,BASROM75);

  /* --- system RAM --- */

  ptrmem=farmalloc(0x20L+0x400L*RAMSIZE);

  if (ptrmem==NULL) {
    fin_ecran();
    fprintf(stderr,"Emu75: No enough memory\n");
    exit(0);
  }
  /* security */
  if (FP_OFF(ptrmem)>15) {
    fin_ecran();
    fprintf(stderr,"Emu75: invalid memory pointer\n");
    exit(0);
  }
  /* ajust for beginning of page */
  seg=FP_SEG(ptrmem)+1;
  tab_seg[4]=seg;
  tab_seg[5]=tab_seg[4]+512;
  tab_seg[6]=tab_seg[5]+512;


  init_ram();

  init_hpil();


  /* allocate memory segments, and read image files */
  nbre=0;
  while (device75[nbre].type) {
    ptrmem=farmalloc(1024L*device75[nbre].taille+32);
    if (ptrmem==NULL) {
      fin_ecran();
      fprintf(stderr,"Emu75: No enough memory for %s\n",
                     device75[nbre].nomfic);
      exit(0);
    }
    /* security */
    if (FP_OFF(ptrmem)>15) {
      fin_ecran();
      fprintf(stderr,"Emu75: invalid memory pointer\n");
      exit(0);
    }
    /* ajust for beginning of page */
    device75[nbre].seg=FP_SEG(ptrmem)+1;

    loadmodule(device75[nbre].seg, device75[nbre].taille,
               device75[nbre].nomfic);

    nbre++;

  }


  /* define the slot addresses */
  /* in this version, slots are allocated in numerical order,
     this is different from the actual HP75 */
  i=0;     // slot counter
  n=0x30;  // 1st slot
  nbre=0;
  fdiag=0;
  while (device75[nbre].type) {
    if (device75[nbre].type!='R') {
      if ( (device75[nbre].nuport==0) &&
           (strnicmp(device75[nbre].nomfic,"ALT",3)==0) ) {
        rombank[i].seg=device75[nbre].seg;
        rombank[i].sltaddr=0x46;
        device75[nbre].sltaddr=0x46;
        tab_seg[3]=rombank[i].seg;  // default to altrom
        i++;
      }
      else if ( (device75[nbre].nuport==0) &&
           (strnicmp(device75[nbre].nomfic,"MEL",3)==0) ) {
        rombank[i].seg=device75[nbre].seg;
        rombank[i].sltaddr=0x4e;
        device75[nbre].sltaddr=0x4e;
        i++;
      }
      else if ( (device75[nbre].nuport==0) &&
           (strnicmp(device75[nbre].nomfic,"DIA",3)==0) ) {
        fdiag=1;  // v1.1
        rombank[i].seg=device75[nbre].seg;
        rombank[i].sltaddr=0x48;    // diag slot
        device75[nbre].sltaddr=0x48;
        i++;
      }
      else {
        if ((i+device75[nbre].taille/8)<20) {
          device75[nbre].sltaddr=n;
          rombank[i].seg=device75[nbre].seg;
          rombank[i].sltaddr=n; n+=2;
          if (device75[nbre].type=='A') rombank[i].fwrit=1;
          i++;
          if (n==0x46) n+=4;
          if (n==0x4e) n+=2;
          if (device75[nbre].taille>8) {
            rombank[i].seg=rombank[i-1].seg+512;
            rombank[i].sltaddr=n; n+=2;
            rombank[i].fwrit=rombank[i-1].fwrit;
            i++;
            if (n==0x46) n+=4;
            if (n==0x4e) n+=2;
          }
          if (device75[nbre].taille>16) {
            rombank[i].seg=rombank[i-1].seg+512;
            rombank[i].sltaddr=n; n+=2;
            rombank[i].fwrit=rombank[i-1].fwrit;
            i++;
            if (n==0x46) n+=4;
            if (n==0x4e) n+=2;
          }
          if (device75[nbre].taille>24) {
            rombank[i].seg=rombank[i-1].seg+512;
            rombank[i].sltaddr=n; n+=2;
            rombank[i].fwrit=rombank[i-1].fwrit;
            i++;
            if (n==0x46) n+=4;
            if (n==0x4e) n+=2;
          }
        }
      }

    }
    nbre++;
  } // end while


  // allocate XMEM memory
  if (instmemic) {
    if (instmemic==2) {
      if (xmemsize==0) xmemsize=32;
      if (xmemsize>256) xmemsize=256;
      n=((xmemsize+31)/32)*32;  // round to 32k crunches
    }
    else {
      if (xmemsize==0) xmemsize=512;
      if (xmemsize>16000) xmemsize=16000;
      n=128; // 128 k imem in RAM with memic1
    }
    ptrmem=farmalloc(0x20L+0x400L*n);
    if (ptrmem==NULL) {
      fin_ecran();
      fprintf(stderr,"Emu75: No enough memory for xmem\n");
      exit(0);
    }
    /* security */
    if (FP_OFF(ptrmem)>15) {
      fin_ecran();
      fprintf(stderr,"Emu75: invalid memory pointer\n");
      exit(0);
    }
    /* ajust for beginning of page */
    seg=FP_SEG(ptrmem)+1;
    /* init memory */
    for (i=0;i<n;i++) {
      ptrmem=MK_FP(seg+i*0x40,0);
      _fmemset(ptrmem,0,0x400);
    }
    /* fill segment table */
    if (instmemic==1) {
      // 128k IMEM
      xmem_seg[0]=seg;
      xmem_seg[1]=xmem_seg[0]+2048;
      xmem_seg[2]=xmem_seg[1]+2048;
      xmem_seg[3]=xmem_seg[2]+2048;
    }
    else {
       // memic 2
       if (n<=128) {
        // XMEM only
        xmem_seg[4]=seg;
        if (n>32)  xmem_seg[5]=xmem_seg[4]+2048;
        if (n>64)  xmem_seg[6]=xmem_seg[5]+2048;
        if (n>96)  xmem_seg[7]=xmem_seg[6]+2048;
      }
      else {
        // 128k IMEM
        xmem_seg[0]=seg;
        xmem_seg[1]=xmem_seg[0]+2048;
        xmem_seg[2]=xmem_seg[1]+2048;
        xmem_seg[3]=xmem_seg[2]+2048;
        // then XMEM up to 128k
        xmem_seg[4]=xmem_seg[3]+2048;
        if (n>160) xmem_seg[5]=xmem_seg[4]+2048;
        if (n>196) xmem_seg[6]=xmem_seg[5]+2048;
        if (n>224) xmem_seg[7]=xmem_seg[6]+2048;
      }
    }
  }

  /*  init processor state and I/O */
  init_register();

}




/* ************************************************************* */
/* save/restore of system state 				 */
/* ************************************************************* */


/* ****************************************** */
/* save_state()                               */
/*                                            */
/* save state of processor and RAM I/O        */
/*   in sys71.dat,                            */
/* RAM in ram75.bin, and PMS RAM in files     */
/* ****************************************** */
void save_state(void)
{
  FILE *fic;
  unsigned char far *ptr;
  int i, j;



  /* sys75.dat */
  fic=fopen("sys75.dat","wb");
  if (fic==NULL) {
    fin_ecran();
    printf("Emu75: error writing sys75.dat\n");
    exit(0);
  }
  // save comp in ff88
  ptr=MK_FP75(0xff88);
  *ptr++ = tcomp;
  *ptr++ = tcomp >>8;
  *ptr++ = tcomp >>16;
  *ptr++ = tcomp >>24;
  // save timer in ff8c
//  ptr=MK_FP75(0xff8c);
  *ptr++ = timer;
  *ptr++ = timer >>8;
  *ptr++ = timer >>16;
  *ptr++ = timer >>24;
  /* I/O : ff00 - ffff*/
  ptr=MK_FP(tab_seg[7],0x1f00);
  _fmemcpy(buffer,ptr,256);
  fwrite(buffer,1,256,fic);
  fclose(fic);


  /* RAM */
  fic=fopen("ram75.dat","wb");
  if (fic==NULL) {
    fin_ecran();
    printf("Emu75: error writing ram75.dat\n");
    exit(0);
  }
  /* RAM 8000 - dfff*/
  ptr=MK_FP(tab_seg[4],0);
  for (i=0;i<24;i++) {
    _fmemcpy(buffer,ptr,1024);
    fwrite(buffer,1,1024,fic);
    ptr+=1024;
  }
  fclose(fic);

  /* PMS RAM */
  i=0;
  while (device75[i].type) {
    if (device75[i].type=='A') {
      fic=fopen(device75[i].nomfic,"wb");
      if (fic==NULL) {
        fin_ecran();
        printf("Emu75: error writing %s\n",device75[i].nomfic);
        exit(0);
      }

      ptr=MK_FP(device75[i].seg,0);
      for (j=0;j<device75[i].taille;j++) {
        _fmemcpy(buffer,ptr,1024);
        fwrite(buffer,1,1024,fic);
        ptr += 1024;
      }

      fclose(fic);
    }
    i++;
  }

  /* XMEM */
  pod_close();  // security
  if (instmemic) {
    fic=fopen(xmemfile,"r+b");
    if (fic==NULL) {
      fic=fopen(xmemfile,"wb");
    }
    if (fic==NULL) {
      fin_ecran();
      printf("Emu75: error writing %s\n",xmemfile);
      exit(0);
    }
    for (i=0;i<8;i++) {
      if (xmem_seg[i]!=0) {
        fseek(fic,i*32768L,0);
        ptr=MK_FP(xmem_seg[i],0);
        for (j=0;j<32;j++) {
          _fmemcpy(buffer,ptr,1024);
          fwrite(buffer,1,1024,fic);
          ptr += 1024;
        }
      }
    }
    fclose(fic);
  }


}


/* ****************************************** */
/* load_state()                               */
/*                                            */
/* restore processor state and RAM I/O        */
/* from sys75.dat                             */
/* ****************************************** */
void load_state(void)
{
  FILE *fic;
  unsigned char far *ptr;
  int i, j;
  char buf[80];

  /* sys75.dat */
  fic=fopen("sys75.dat","rb");
  if (fic==NULL) {
    sprintf(buf,"Emu75: error reading sys75.dat\r\n");
    disp_string(buf);
  }
  else {

    /* I/O : ff00 - ffff*/
    ptr=MK_FP(tab_seg[7],0x1f00);
    fread(buffer,1,0x0100,fic);
    _fmemcpy(ptr,buffer,0x100);
    fclose(fic);

    // restore comp from ff88
    ptr=MK_FP75(0xff8b);
    tcomp  = *ptr--;  tcomp<<=8;
    tcomp += *ptr--;  tcomp<<=8;
    tcomp += *ptr--;  tcomp<<=8;
    tcomp += *ptr--;
    // restore timer from ff8c
    ptr=MK_FP75(0xff8f);
    timer  = *ptr--;  timer<<=8;
    timer += *ptr--;  timer<<=8;
    timer += *ptr--;  timer<<=8;
    timer += *ptr;

    // v1.2: restore compstat
    ptr=MK_FP75(0xff80);
    compstat=*ptr;

    // clr wakeup bit from comp
    ptr=MK_FP75(0xff02);
    i=*ptr;
    *ptr = (i & 0xef);

    // power status
    ptr=MK_FP75(0xff82);
    i=*ptr;
    *ptr = (i & 0xbf) | 0x20;  // fully awake


    /* RAM */
    fic=fopen("ram75.dat","rb");
    if (fic==NULL) {
      sprintf(buf,"Emu75: error reading ram75.dat\r\n");
    }
    else {
      /* RAM 8000 - dfff*/
      ptr=MK_FP(tab_seg[4],0);
      for (i=0;i<24;i++) {
        fread(buffer,1,0x0400,fic);
        _fmemcpy(ptr,buffer,0x400);
        ptr+=0x400;
      }
      fclose(fic);
    }
  }

  /* XMEM */
  if (instmemic) {
    fic=fopen(xmemfile,"rb");
    if (fic==NULL) {
      sprintf(buf,"Emu75: error reading %s\r\n",xmemfile);
      disp_string(buf);
    }
    else {
      for (i=0;i<8;i++) {
        if (xmem_seg[i]!=0) {
          fseek(fic,i*32768L,0);
          ptr=MK_FP(xmem_seg[i],0);
          for (j=0;j<32;j++) {
            fread(buffer,1,1024,fic);
            _fmemcpy(ptr,buffer,1024);
            ptr += 1024;
          }
        }
      }
      fclose(fic);
    }
  }
}



/* ****************************************** */
/* checkrom()                                 */
/*                                            */
/* check the checksum of ROMs                 */
/* ****************************************** */
void checkrom(void)
{
  int i, j, n, m;
  unsigned char far *ptr;
  unsigned char *ptr1;
  char nom[40];
  struct portdevice *deviceptr;

  /* test internal ROM */
  ptr=MK_FP(tab_seg[0],0);
  strcpy(nom,SYSROM75);
  ptr1=strchr(nom,'.');
  if (ptr1) *ptr1=0;
  strupr(nom);
  printf("00 %-8s",nom);
  for (i=0;i<3;i++) {
    n=0;
    printf(" %d",i+1);
    for (j=0;j<8192;j++) {
      n += *ptr++;
      if (n>255) {
	n++;
	n &= 255;
      }
    }
    if (n==255) printf("G"); else printf("B(%x)",n);
  }
  if (BASROM75[0]!=0) {
    printf("\n");
    strcpy(nom,BASROM75);
    ptr1=strchr(nom,'.');
    if (ptr1) *ptr1=0;
    strupr(nom);
    printf("01 %-8s",nom);
    i=0;
  }
  ptr=MK_FP(tab_seg[7],0);
  n=0;
  printf(" %d",i+1);
  for (j=0;j<(8192-256);j++) {
    n += *ptr++;
    if (n>255) {
      n++;
      n &= 255;
    }
  }
  if (n==255) printf("G"); else printf("B(%x)",n);
  printf("\n");

  /* test switched ROM */
  deviceptr = &device75[0];
  for (i=0;i<16;i++) {
    if (deviceptr->type==0) break;  /* end of devices */
    if (deviceptr->type!='R') {  /* ROM/RAM modules */
      strcpy(nom,deviceptr->nomfic);
      ptr1=strchr(nom,'.');
      if (ptr1) *ptr1=0;
      strupr(nom);
      printf("%02X %-8s",deviceptr->sltaddr,nom);
      m=deviceptr->taille/8;
      ptr=MK_FP(deviceptr->seg,0);
      for (i=0;i<m;i++) {
	n=0;
	printf(" %d",i+1);
	for (j=0;j<8192;j++) {
	  n += *ptr++;
	  if (n>255) {
	    n++;
	    n &= 255;
	  }
	}
	if (n==255) printf("G"); else printf("B(%x)",n);
      }
      printf("\n");
    }
    deviceptr++;
  }

}





/* ************************************************************* */
/* Emulation							 */
/* ************************************************************* */





/* ****************************************** */
/* simulator()                                */
/*                                            */
/* simulator high level main loop             */
/* returns 1 if ended by deep sleep,          */
/*   or 2 if call to debugger                 */
/* ****************************************** */
int simulator(void)
{
  int n, k;
  long t, t0;
  int ret;

  ret=0;
  do {
    n=executeCapricorn(speed_emu);
    /* returns 0 if OK,                           */
    /*   -1 if invalid, -2 if break               */
    /*    1 if light sleep, 2 if deep sleep */
    if (n) {
      pod_close(); // 1.3b
      if (n==2) {
       /* deep sleep */
       ret=1;
       break;
      }
      else if (n==1) {
	traite_display(0);
	/* wait for key or timer */
	n=0;
        if (!flowpower) {
          do {
            k=traite_touche(0);
            if (k<0) ret=2;  /* call to debug */
//            clock();
            if (tsttimer()) {
              incr_timer(900);
              n=check_timer();
            }
            send_buffer();
#if 0
	    if (f_xil) {
              if (hpil_ctrl()==0) traite_extil();
            }
            if (hpil_ctrl()==0) traite_extil2();   /* added JFG 14/04/08 */

#endif
          } while ((k==0)&&(n==0));
        }
        else {
          t=clock()*UNIT_TICKS;
          do {
            k=traite_touche(1);
          } while (k==0);
   	  if (k<0) ret=2;  /* call to debug */
          tsttimer();
          t0=clock()*UNIT_TICKS;
          t=t0-t;  /* elapsed time in ms */
          t=t*16 ; /* ticks to adjust */
          incr_timer(t);
        }
      }
      else if (n<0)
	ret=2;   /* break point or illegal */
    }


    if (tsttimer()) {
      send_buffer();
      // HP75: timer in 2^-14 s units = 1/16384 = 61us.
      incr_timer(900);  /* about 55 ms / (1/16384 s)   */
      check_timer();
      traite_display(1);
      k=traite_touche(0);
      if (k<0) ret=2;
    }

  } while (ret==0);

  printf("\n");
  restore_cursor();

  return(ret);
}








/* ****************************************** */
/* usage()                                    */
/*                                            */
/* ****************************************** */
void static usage(void)
{
  printf("Emu75: emulates the HP-75C machine\n");
  printf("Syntax:  EMU75 [options]\n");
  printf("Options:\n");
  printf("         /d : call debugger at startup\n");
  printf("         /m : HP-IL frame monitor\n");
//  printf("         /t : trace system instructions\n");
//  printf("         /s : slow host (i.e. 286 or HP95LX)\n");
  printf("         /f : fast host (native 16/32-bit OS)\n");
  printf("V1.3   (C) 2010-2018 J-F Garnier   http://www.jeffcalc.hp41.eu/emu75\n\n");
  // added for open source release:
  printf("This program is free software; you can redistribute it and/or\n");
  printf("modify it under the terms of the GNU General Public License\n");
  printf("as published by the Free Software Foundation; either version 2\n");
  printf("of the License, or (at your option) any later version.\n\n");
  printf("This program is distributed in the hope that it will be useful,\n");
  printf("but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
  printf("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
  printf("GNU General Public License for more details.\n\n");
  printf("You should have received a copy of the GNU General Public License\n");
  printf("along with this program; if not, write to the Free Software\n");
  printf("Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.\n");
  exit(0);
}



/* ****************************************** */
/* main()                                     */
/*                                            */
/* Emu75 main entry                           */
/*   read and set options                     */
/* ****************************************** */

void main(int argc, char *argv[])
{
  int i, n, ret;
  char *arg;
  int dbg;

  _directvideo=0;    /* BIOS call */

  f_trace=0;
  fmonit=0;
  speed_emu=128;  // equivalent to option /s2
  flowpower=0;
  dbg=0;

  /* read options */
  for (i=1;i<argc;i++) {
    arg=argv[i];
    if (arg[0]=='/') {
      switch(arg[1]) {
	case 'd':
//	  init_label("ENTR75");
	  dbg=1;
          break;
	case 'l':
	  flowpower=1;
	  break;
	case 'm':
	  n=atoi(&arg[2]);
          if (n==0) n=1;
	  fmonit=n;
          init_log();
	  break;
	case 't':
	  n=atoi(&arg[2]);
	  if (n>0) f_trace=n; else f_trace=1;
	  break;
	case 's':
	  n=atoi(&arg[2]);
          if (n==0) speed_emu=255;
	  if (n==1) speed_emu=192;
	  if (n>1)  speed_emu=256/n;
          _directvideo=1;
          break;
	case 'f':
	  n=atoi(&arg[2]);
          if (n==0) n=4;
	  if (n>=1) speed_emu=256*n;
	  break;
	case '?':
	  usage();
	  break;
	default:
	  fprintf(stderr,"Emu71: unknown option '%c'\n",arg[1]);
	  exit(0);
      }
    }
  }

  init_label("ENTR75");

  prepa_ecran(0);

  init_emulate(dbg);

  load_state();
  maj_date();   // 0.95: update the RTC

  if (dbg)
    ret=commande();
  else {
    ret=0;
    ret=simulator(); /* ret=1 if deep sleep , 2 if break */
    if (ret!=1)
      ret=commande(); /* ret=1 if deep sleep, -1 if exit */
  }
  if (ret==1) save_state();

  EndEms();
  fin_ecran();
}

